#ifndef _MAZE_H
#define _MAZE_H

#include <fstream>
#include <string>
#include <stdio.h>
#include <vector>
using namespace std;

#include "Grid.h"

void SoundDraw();

class Maze
{
public:

	int dim_x, dim_y;								// Dimension of the maze
	int numTiles;									// Number of tile files
	int numSounds;									// Number of sound files
	int difficulty;									// Difficulty setting (not currently used)

	Grid *tGrid;									// Maze Grid
	vector<Tile> tTiles;							// List of the tile data

	FMOD::System *fsystem;							// FMOD system
	FMOD::Sound *fsounds[9];						// List of sound data

	FMOD::Geometry *fgeometry;						// FMOD geometry object

	int st_x, st_z, st_subt_x, st_subt_z;			// start point
	int ed_x, ed_z, ed_subt_x, ed_subt_z;			// end point (not currently used)

	struct SoundObject								// Soundobject class
	{
		int tile_x;									// Tile location of object
		int tile_z;
		int pos_x;									// Tile index position of 3x3 subgrid within tile
		int pos_z;
		float abs_x;								// Absolute world position of object
		float abs_z;
		bool hidden;								// Has the tile been hidden?
		float intensity;							// Sound volume
		int soundID;								// ID of sound in fsounds to be used for this object
		FMOD::Channel *channel;						// FMOD channel for the sound
	};
	int num_objects;								// Number of soundobjects in the maze
	vector<SoundObject> fsound_objects;				// List of soundobjects
	
	Maze();
	
	void setSize(int x, int y);
	bool LoadTiles(char *dir);
	bool LoadSounds(char *dir);
	void LoadTileTextures();
	void DrawMaze(int current_x, int current_z);
	bool LoadMaze(char *filename);
};

Maze::Maze()
{
	dim_x = 1; dim_y = 1;
	numTiles = 0;
	numSounds = 0;
	difficulty = 0;

	FMOD_RESULT result;
	result = FMOD::System_Create(&fsystem);		// Create the main FMOD system object.
	if(result != FMOD_OK)
	{
	    MessageBox(NULL,"FMOD error creating!","Error",MB_OK);
	}
	
	result = fsystem->init(100, FMOD_INIT_NORMAL, 0);	// Initialize FMOD.
	if(result != FMOD_OK)
	{	
		MessageBox(NULL,"FMOD error initializing!","Error",MB_OK);	
	}
}

void Maze::setSize(int x, int y)			// Set the size of the maze
{
	dim_x = x; dim_y = y;
}

void Maze::DrawMaze(int current_x, int current_y)		// Draw the maze and soundobjects
{
	glPushMatrix();
	for(int y = 0; y < dim_y; y++)
	{
		glPushMatrix();
		for(int x = 0; x < dim_x; x++)
		{
			if(y>=(current_y-2) && y<=(current_y+2))
			{
				if(x>=(current_x-2) && x<=(current_x+2))
				{
					glPushMatrix();
					glRotatef((GLfloat)((tGrid->getTileRot(x,y)-1)*-90), 0.0f, 1.0f, 0.0f);
					tTiles[tGrid->getTileID(x,y)-1].Draw();
					glPopMatrix();
				}
			}
			glTranslatef(39.0f, 0.0f, 0.0f);		// tiles are 39x39 (subtiles are 13x13)
		}
		glPopMatrix();
		glTranslatef(0.0f, 0.0f, 39.0f);
	}
	glPopMatrix();

	for(int i = 0; i < num_objects; i++)
	{
		if(!fsound_objects[i].hidden)
		{
			glPushMatrix();
			int x_t = fsound_objects[i].tile_x-1;
			int z_t = fsound_objects[i].tile_z-1;
			int x_p = fsound_objects[i].pos_x-1;
			int z_p = fsound_objects[i].pos_z-1;
			glTranslatef(x_t*39.0f+x_p*13.0f-12.0f, 5.0f, z_t*39.0f+z_p*13.0f-12.0f);
			SoundDraw();
			glPopMatrix();
		}
	}
}

bool Maze::LoadTiles(char *dir)
{
	char *input = new char[256];
	char *filename = new char[256];
	strcpy(filename, dir);
	strcat(filename, "info.txt");
	ifstream inputFile(filename, ios::in | ios::binary);
	if(inputFile.fail())
		return false;
	inputFile.getline(input,256,' ');
	numTiles = atoi(input);

	for(int x = 0; x < numTiles; x++)
	{
		strcpy(filename, dir);
		char *num = new char[5];
		_itoa_s(x+1, num, 5, 10);
		if(x<10) { strcat(filename, "0"); }
		strcat(filename, num);
		strcat(filename, ".ms3d");
		Tile temp;
		if(!temp.LoadData(filename)) 
		{
			MessageBox(NULL,"Tile Loading Failure","ERROR",MB_OK|MB_ICONEXCLAMATION);
			return false;
		}
		tTiles.push_back(temp);
	}

	inputFile.close();

	return true;
}

bool Maze::LoadSounds(char *dir)
{
	char *input = new char[256];
	char *filename = new char[256];
	strcpy(filename, dir);
	strcat(filename, "info.txt");
	ifstream inputFile(filename, ios::in | ios::binary);
	if(inputFile.fail())
		return false;
	inputFile.getline(input,256,' ');
	numSounds = atoi(input);
	inputFile.close();

	FMOD_RESULT result;

	for(int x = 0; x < numSounds; x++)
	{
		strcpy(filename, dir);
		char *num = new char[5];
		_itoa_s(x+1, num, 5, 10);
		if(x<10) { strcat(filename, "0"); }
		strcat(filename, num);
		strcat(filename, ".wav");
		
		result = fsystem->createSound(filename, FMOD_HARDWARE | FMOD_3D, 0, &fsounds[x]);
		if(result != FMOD_OK)
			return false;
	    result = fsounds[x]->set3DMinMaxDistance(30.0f, 10000.0f);
	}

	return true;
}

void Maze::LoadTileTextures()
{
	for(int c = 0; c < numTiles; c++)
	{
		tTiles[c].LoadTextures();
	}
}

bool Maze::LoadMaze(char *filename)
{
	char *tileset = new char[50];
	char *soundset = new char[50];
	char *ptr = new char[10];
	FMOD_RESULT result;

	ifstream inputFile(filename, ios::in | ios::binary);
	if (inputFile.fail())
		return false;

	inputFile.getline(tileset, 50);
	strtok(tileset, ":");
	if(!LoadTiles(tileset))
	{
		MessageBox(NULL,"Loading Tiles Failed.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return false;
	}

	inputFile.getline(soundset, 50);
	strtok(soundset, ":");
	if(!LoadSounds(soundset))
	{
		MessageBox(NULL,"Loading Sounds Failed.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return false;
	}

	inputFile.getline(ptr, 10);
	inputFile.getline(ptr, 10);
	char *tok = strtok(ptr, " ");
	dim_x = atoi(tok);
	tok = strtok(NULL, " ");
	dim_y = atoi(tok);
	tGrid = new Grid(dim_x, dim_y);

	inputFile.getline(ptr, 10);
	inputFile.getline(ptr, 10);
	char *sttok = strtok(ptr, " ");
	st_x = atoi(sttok);
	sttok = strtok(NULL, " ");
	st_z = atoi(sttok);
	sttok = strtok(NULL, " ");
	st_subt_x = atoi(sttok);
	sttok = strtok(NULL, " ");
	st_subt_z = atoi(sttok);
	inputFile.getline(ptr, 10);
	sttok = strtok(ptr, " ");
	ed_x = atoi(sttok);
	sttok = strtok(NULL, " ");
	ed_z = atoi(sttok);
	sttok = strtok(NULL, " ");
	ed_subt_x = atoi(sttok);
	sttok = strtok(NULL, " ");
	ed_subt_z = atoi(sttok);

	inputFile.getline(ptr, 10);
	inputFile.getline(ptr, 10);
	sttok = strtok(ptr, " ");
	num_objects = atoi(sttok);
	for(int w = 0; w < num_objects; w++)
	{
		SoundObject temp;

		inputFile.getline(ptr, 10);
		sttok = strtok(ptr, " ");
		temp.tile_x = atoi(sttok);
		sttok = strtok(NULL, " ");
		temp.tile_z = atoi(sttok);
		sttok = strtok(NULL, " ");
		temp.pos_x = atoi(sttok);
		sttok = strtok(NULL, " ");
		temp.pos_z = atoi(sttok);

		temp.abs_x = (temp.tile_x-1)*39.0f+(temp.pos_x-1)*13.0f-12.0f;
		temp.abs_z = (temp.tile_z-1)*39.0f+(temp.pos_z-1)*13.0f-12.0f;
		temp.intensity = 1.0f;
		temp.soundID = w%4+5;
		temp.channel = 0;
		temp.hidden = false;
		fsound_objects.push_back(temp);
	}

	int id, rot = 0;
	int triangle_count = 0;
	int vertex_count = 0;

	for(int y = 0; y < dim_y; y++)
	{
		inputFile.getline(ptr, 10);
		for(int x = 0; x < dim_x; x++)
		{
			inputFile.getline(ptr, 10);
			tok = strtok(ptr, " ");
			id = atoi(tok);
			tok = strtok(NULL, " ");
			rot = atoi(tok);
			triangle_count += tTiles[id-1].tModel->m_numTriangles;
			vertex_count += tTiles[id-1].tModel->m_numVertices;
			tGrid->setTileID(id,x,y);
			tGrid->setTileRot(rot,x,y);
		}
	}

	// ** Construct FMOD Geometry **

	result = fsystem->setGeometrySettings(max(dim_x, dim_y)*39.0f+100.0f);
	if(result != FMOD_OK)
	    MessageBox(NULL,"Could not create FMOD geometry settings.","Error",MB_OK);
	result = fsystem->createGeometry(triangle_count+10, 100000, &fgeometry);
	if(result != FMOD_OK)
	    MessageBox(NULL,"Could not create FMOD geometry.","Error",MB_OK);

	int polyindex = 0;

	for(int v = 0; v < dim_y; v++)
	{
		for(int u = 0; u < dim_x; u++)
		{
			int id = tGrid->getTileID(u,v)-1;
			int rotation = tGrid->getTileRot(u,v);
			int numverts = 3;
			int num_triangles = tTiles[id].tModel->m_numTriangles;

			for(int t = 0; t < num_triangles; t++)
			{
				FMOD_VECTOR vertices[3];
				for (int i = 0; i < 3; i++)
				{
					vertices[i].x = tTiles[id].tModel->m_pVertices[tTiles[id].tModel->m_pTriangles[t].m_vertexIndices[i]].m_location[0];
					vertices[i].y = tTiles[id].tModel->m_pVertices[tTiles[id].tModel->m_pTriangles[t].m_vertexIndices[i]].m_location[1];
					vertices[i].z = tTiles[id].tModel->m_pVertices[tTiles[id].tModel->m_pTriangles[t].m_vertexIndices[i]].m_location[2];

					float temp = 0.0f;
					if(rotation==2) { temp = vertices[i].x; vertices[i].x = vertices[i].z; vertices[i].z = temp; vertices[i].x = -vertices[i].x; }
					if(rotation==3) { vertices[i].x = -vertices[i].x; vertices[i].z = -vertices[i].z; }
					if(rotation==4) { temp = vertices[i].x; vertices[i].x = vertices[i].z; vertices[i].z = temp; vertices[i].z = -vertices[i].z; }

					vertices[i].x += u*39.0f;
					vertices[i].z += v*39.0f;
				}
					
				result = fgeometry->addPolygon(0.5f, 0.5f, false, numverts, vertices, &polyindex);
				if(result != FMOD_OK)
				{
					MessageBox(NULL,FMOD_ErrorString(result),"Error",MB_OK);
					exit(-1);
				}
			}
		}
	}

	inputFile.close();
	return true;
}

void SoundDraw()
{
	glBegin(GL_QUADS);
		glVertex3f( 1.0f, 1.0f,-1.0f);
		glVertex3f(-1.0f, 1.0f,-1.0f);
		glVertex3f(-1.0f, 1.0f, 1.0f);
		glVertex3f( 1.0f, 1.0f, 1.0f);

		glVertex3f( 1.0f,-1.0f, 1.0f);
		glVertex3f(-1.0f,-1.0f, 1.0f);
		glVertex3f(-1.0f,-1.0f,-1.0f);
		glVertex3f( 1.0f,-1.0f,-1.0f);

		glVertex3f( 1.0f, 1.0f, 1.0f);
		glVertex3f(-1.0f, 1.0f, 1.0f);
		glVertex3f(-1.0f,-1.0f, 1.0f);
		glVertex3f( 1.0f,-1.0f, 1.0f);

		glVertex3f( 1.0f,-1.0f,-1.0f);
		glVertex3f(-1.0f,-1.0f,-1.0f);
		glVertex3f(-1.0f, 1.0f,-1.0f);
		glVertex3f( 1.0f, 1.0f,-1.0f);

		glVertex3f(-1.0f, 1.0f, 1.0f);
		glVertex3f(-1.0f, 1.0f,-1.0f);
		glVertex3f(-1.0f,-1.0f,-1.0f);
		glVertex3f(-1.0f,-1.0f, 1.0f);

		glVertex3f( 1.0f, 1.0f,-1.0f);
		glVertex3f( 1.0f, 1.0f, 1.0f);
		glVertex3f( 1.0f,-1.0f, 1.0f);
		glVertex3f( 1.0f,-1.0f,-1.0f);
	glEnd();
}
#endif